/*
** Module   : DirtyTempTableHelper.java
** Abstract : Helper class which generates temp table DDL
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 CA  20080424   @38110 Created initial version. Generates DDL for
**                           global temp tables with global indexes.
** 002 CA  20080430   @38158 Changes due to incorrect index creation: the
**                           indexes must be retrieved from the "main" 
**                           database. Added method to build the required
**                           "dirty" table and index.
** 003 ECF 20080522   @38603 Relocated createTable() helper method. A
**                           modified version of this method now resides
**                           in DefaultDirtyShareManager.
** 004 ECF 20080611   @38694 Changed invocation of queryIndexData(). We
**                           now use a common implementation in the
**                           Persistence class.
** 005 ECF 20090324   @42124 Set all dirty indices to be non-unique. It is
**                           possible for multiple records to be created
**                           simultaneously in different sessions whose
**                           default values would normally violate a unique
**                           index; however, until that index is explicitly
**                           updated, the default values must not violate the
**                           unique constraint in the dirty database.
** 006 ECF 20090511   @42152 Fixed create index statement generation. Avoid
**                           creating a superfluous index on the primary key.
**                           Avoid tacking an additional primary key component
**                           onto the end of indexes.
** 007 CA  20120710          During server startup, getIndexes will hide any
**                           persistence errors, as the terminal is not 
**                           configed yet - so, we need to explicitly log any
**                           persistence errors.
** 008 ECF 20140318          Reduced amount of index metadata queried from database to optimize
**                           performance. Replaced Apache commons logging with J2SE logging.
** 009 ECF 20140822          Added implementation of isNotNullAllowed.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.logging.*;
import org.hibernate.mapping.Table;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * Helper class which generates and caches SQL phrases to create and drop
 * temporary tables, and to create indexes in those tables.
 * <p>
 * This class is responsible for global temporary tables support: the created
 * tables and indices will be global and available to all sessions.
 */
public final class DirtyTempTableHelper
extends TempTableHelper
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(DirtyTempTableHelper.class);

   /** 
    * Cache with the indexes retrieved from the database, for each table and
    * database.
    */
   private static final Map<Database, Map<String, List<P2JIndex>>> indexCache = new HashMap<>();
   
   /**
    * Constructor.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   database
    *          The database for which we generate the sql's.
    */
   DirtyTempTableHelper(P2JDialect dialect, Class<?> dmoIface, Database database)
   {
      super(dialect, dmoIface, database);
   }

   /**
    * This method will return the needed create table string, which is the 
    * "create temporary table" string.
    * 
    * @return  The create table string.
    */
   protected String getCreateTableStatement()
   {
      return "create temporary table ";
   }
   
   /**
    * This method returns an empty string: the indexes must be global.
    * 
    * @return An empty string.
    */
   protected String getAdditionalIndexSuffix()
   {
      return "";
   }

   /**
    * This method is overridden to return an empty list - once created, the
    * indices can not be dropped. 
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  An empty list.
    */
   protected List<String> generateSQLDropIndexList(P2JDialect dialect,
                                                   Class<?> dmoIface, 
                                                   Table primary, 
                                                   List<Table> secondaries)
   {
      return Collections.emptyList();
   }
   
   /**
    * Generate the dialect-specific SQL statement used to create an index
    * for the temp table.
    * <p>
    * Override the parent's implementation to never add the primary key as the
    * last component of the index.
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   index
    *          P2J index descriptor.
    * @param   table
    *          Hibernate database table helper.
    * @param   mpid
    *          <code>true</code> to make multiplex ID the leading index
    *          component;  <code>false</code> to omit it. The multiplex ID 
    *          column will not be indexed if the {@link #schema} is not the
    *          {@link DatabaseManager#TEMP_TABLE_SCHEMA}.
    * @param   overridePK
    *          If <code>true</code>, do not add primary key column to a
    *          non-unique index;  if <code>false</code>, add primary key
    *          column to a non-unique index.  This method overrides the parent
    *          to ignore this parameter and always pass <code>true</code> to
    *          the parent's implementation.
    *
    * @return  DDL statement text.
    */
   protected String generateSQLCreateIndex(P2JDialect dialect,
                                           P2JIndex index,
                                           Table table,
                                           boolean mpid,
                                           boolean overridePK)
   {
      return super.generateSQLCreateIndex(dialect, index, table, mpid, true);
   }
   
   /**
    * This method is overridden to return an empty list - once created, the
    * tables can not be dropped. 
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  An empty list.
    */
   @Override
   protected List<String> generateSQLDropTableList(P2JDialect dialect,
                                                   Table primary, 
                                                   List<Table> secondaries)
   {
      return Collections.emptyList();
   }

   /**
    * This method will retrieve the computed columns for the required DMO, as
    * found in the indexes retrieved from the database.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the computed columns.
    * 
    * @return  A map with the computed columns associated with this DMO.
    */
   @Override
   protected Map<String, P2JIndexComponent> getComputedCols(String schema, Class<?> dmoIface)
   {
      Map<String, P2JIndexComponent> computedCols = new HashMap<>();
      
      Iterator<P2JIndex> iter = getIndexes(schema, dmoIface);
      while (iter.hasNext())
      {
         P2JIndex index = iter.next();
         
         Iterator<P2JIndexComponent> iter2 = index.components();
         while (iter2.hasNext())
         {
            P2JIndexComponent idxComp = iter2.next();
            
            if (idxComp.isCharType())
            {
               computedCols.put(idxComp.getName(), idxComp);
            }
         }
      }
      
      return computedCols;
   }

   /**
    * This method will return the indexes associated with a certain DMO, as
    * found in the associated database. It must be noted that those indexes
    * will not contain any foreign-key columns (columns with their name ending
    * with a "__fk" suffix); those columns are removed from the index metadata
    * and if an index contains only a such column, that index will not be 
    * returned.  Furthermore, all indexes will be set to be non-unique, since
    * newly created DMOs can be added from multiple sessions, and the default
    * values of unique properties could conflict.
    * <p>
    * Once retrieved, the indexes will be kept in a {@link #indexCache cache},
    * for each table, by database.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the indexes.
    * 
    * @return  An iterator for the indexes associated with this DMO.
    * 
    * @throws  RuntimeException
    *          if there is any error retrieving the index metadata for the given schema and DMO.
    */
   @Override
   protected Iterator<P2JIndex> getIndexes(String schema, Class<?> dmoIface)
   {
      Database mainDB = new Database(database.getName());
      String table = getTableName(mainDB, dmoIface);
      ArrayList<P2JIndex> indexes = null;
      
      synchronized (indexCache)
      {
         if (indexCache.containsKey(mainDB) == false)
         {
            indexCache.put(mainDB, new HashMap<String, List<P2JIndex>>());
         }
         
         Map<String, List<P2JIndex>> cached = indexCache.get(mainDB);
         
         // check if the indexes for the table have already been read
         if (cached.containsKey(table))
         {
            return cached.get(table).iterator();
         }
         
         // get the index list from the DMOIndex cache (which will query the main database for
         // metadata if there is no cache hit); this will give us shared P2JIndex instances, so we
         // have to make copies to modify them
         try
         {
            Iterator<P2JIndex> indexIter = DMOIndex.databaseIndexes(schema, mainDB, dmoIface);
            indexes = new ArrayList<>();
            
            // copy each index, remove the __fk columns if present, and ensure they are not set to
            // be unique
            
            outer:
            while (indexIter.hasNext())
            {
               P2JIndex index = indexIter.next();
               P2JIndex copy = new P2JIndex(index);
               
               copy.setUnique(false);
               
               Iterator<P2JIndexComponent> compIter = copy.components();
               for (int i = 0; compIter.hasNext(); i++)
               {
                  P2JIndexComponent idxComp = compIter.next();
                  String compName = idxComp.getName();
                  
                  if (compName.endsWith(ORMHandler.FK_SUFFIX))
                  {
                     compIter.remove();
                  }
                  else if (i == 0 &&
                           DatabaseManager.PRIMARY_KEY.equals(compName) &&
                           !compIter.hasNext())
                  {
                     // ignore index on primary key; the dirty database will create such an index
                     // automatically when the table is created.
                     continue outer;
                  }
               }
               
               // if the foreign key column this was the only component for this index, ignore the
               // index
               if (index.components().hasNext() == false)
               {
                  continue outer;
               }
               
               copy.compact();
               indexes.add(copy);
            }
            
            indexes.trimToSize();
         }
         catch (PersistenceException exc)
         {
            String msg = String.format("Error determining indexes for %s (%s, %s)",
                                       dmoIface.getName(), schema, mainDB.toString());
            LOG.log(Level.SEVERE, msg, exc);
            
            // this is pretty much a fatal error, since we will run into problems downstream if
            // this information is unavailable
            throw new RuntimeException(msg, exc);
         }
         
         // cache the found indexes
         cached.put(table, indexes);
      }
      
      return indexes.iterator();
   }
   
   /**
    * Indicate whether non-nullable columns are permitted.
    * 
    * @return  <code>true<code> if non-nullable columns are permitted, else <code>false</code>.
    */
   protected boolean isNotNullAllowed()
   {
      return false;
   }
}
