/*
** Module   : ValidationData.java
** Abstract : Validation parameters
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20090331   @42122 Created initial version. Validation parameters.
** 002 ECF 20090511   @42156 Fixed default constructor. Last two arguments
**                           were transposed.
** 003 ECF 20090623   @43224 Added isTransient flag and getter.
** 004 ECF 20140826          Added deferred flag and getter/setter.
*/

package com.goldencode.p2j.persist;

import java.util.*;

/**
 * This object stores validation parameters to assist the {@link DMOValidator}
 * during record validation.  Depending upon the type of validation requested
 * (e.g., full, remaining, dirty), the parameters will vary.  This object
 * references the names of all dirty properties, the names of all unique and
 * non-unique indexes that should be checked, whether dirty updates should be
 * shared via the dirty share manager, and whether an insert into the dirty
 * share manager is required.
 */
final class ValidationData
{
   /** All dirty properties requiring validation */
   private final Set<String> dirtyProps;
   
   /** All unique indexes requiring validation/sharing */
   private final Collection<String> dirtyUnique;
   
   /** All non-unique indexes requiring sharing */
   private final Collection<String> dirtyNonUnique;
   
   /** Is an insert required in the dirty share manager? */
   private final boolean insertRequired;
   
   /** Should insert/update information be shared with other contexts? */
   private final boolean shareDirty;
   
   /** Is the affected record newly created and still transient? */
   private final boolean isTransient;
   
   /** Was some validation work and dirty database sharing deferred? */
   private boolean deferred = false;
   
   /**
    * Constructor.
    * 
    * @param   dirtyProps
    *          All dirty properties requiring validation.
    * @param   dirtyUnique
    *          All unique indexes requiring validation/sharing.
    * @param   dirtyNonUnique
    *          All non-unique indexes requiring sharing.
    * @param   insertRequired
    *          Is an insert required in the dirty share manager?
    * @param   shareDirty
    *          Should insert/update information be shared with other contexts?
    * @param   isTransient
    *          Is the affected record newly created and still transient?
    */
   ValidationData(Set<String> dirtyProps,
                  Collection<String> dirtyUnique,
                  Collection<String> dirtyNonUnique,
                  boolean insertRequired,
                  boolean shareDirty,
                  boolean isTransient)
   {
      this.dirtyProps = Collections.unmodifiableSet(dirtyProps);
      this.dirtyUnique = dirtyUnique;
      this.dirtyNonUnique = dirtyNonUnique;
      this.insertRequired = insertRequired;
      this.shareDirty = shareDirty;
      this.isTransient = isTransient;
   }
   
   /**
    * Get the set of dirty property names.
    * 
    * @return  Dirty property names.
    */
   Set<String> getDirtyProperties()
   {
      return dirtyProps;
   }
   
   /**
    * Get the collection of dirty, unique, index names.
    * 
    * @return  Collection of dirty, unique, index names, in ascending
    *          alphabetical order.
    */
   Collection<String> getDirtyUnique()
   {
      return dirtyUnique;
   }
   
   /**
    * Get the collection of dirty, non-unique, index names.
    * 
    * @return  Collection of dirty, non-unique, index names, in ascending
    *          alphabetical order.
    */
   Collection<String> getDirtyNonUnique()
   {
      return dirtyNonUnique;
   }
   
   /**
    * Indicate whether we need to insert a record into the dirty share
    * manager to reflect a new record.
    * 
    * @return  <code>true</code> to make the insert, else <code>false</code>.
    */
   boolean isInsertRequired()
   {
      return insertRequired;
   }
   
   /**
    * Indicate whether DMO insert/update information should be shared with
    * other contexts via the dirty share manager.
    * 
    * @return  <code>true</code> to share updates, else <code>false</code>.
    */
   boolean isShareDirty()
   {
      return shareDirty;
   }
   
   /**
    * Indicate whether the affected record is newly created and is still
    * transient.
    * 
    * @return  <code>true</code> if transient, else <code>false</code>.
    */
   boolean isTransient()
   {
      return isTransient;
   }
   
   /**
    * Indicate whether possibly uniqueness validation and dirty sharing were deferred because not
    * all properties in an index were dirty for a transient record. This will require the index
    * to be validated when the record is flushed.
    * 
    * @return  <code>true</code> if deferred, else <code>false</code>.
    */
   boolean wasDeferred()
   {
      return deferred;
   }
   
   /**
    * Set the deferred flag to indicate dirty sharing and possibly uniqueness validation of an
    * updated property were deferred.
    * 
    * @see     #wasDeferred()
    */
   void setDeferred()
   {
      deferred = true;
   }
}
