/*
** Module   : DirtyShareMultiplexer.java
** Abstract : Defines a remote API to manage the information about uncommitted
**            modifications to database records.
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 SVL 20080605   @38558 Created initial version. Defines a remote API
**                           to manage the information about uncommitted
**                           modifications.
** 002 ECF 20080612   @38726 Replaced isEntityDirty() with
**                           getEntityDirtyState(). New method returns
**                           state information (is entity dirty with
**                           changes and/or deletes), rather than a simple
**                           boolean.
** 003 CA  20080716   @39117 Renamed method lockUniqueIndexes to 
**                           lockAllIndexes - this will lock all the 
**                           indexes for a certain entity. Modified the
**                           delete(...) and rollbackInsert(...) methods
**                           so the index locking can be disabled.
** 004 ECF 20080815   @39546 Added support for global change events. These
**                           are events that represent index-changing
**                           inserts, updates, or deletes, which are
**                           processed by sessions other than the session
**                           in which the event originated.
** 005 ECF 20090513   @42171 Modified variant of update() method. Added lock
**                           parameter to make index locking optional.
** 006 ECF 20090602   @42574 Added support for Progress isolation leak quirk.
**                           Added updateSnapshots() method.
** 007 ECF 20090609   @42645 Modified cleanupChanges(). The list of entities
**                           which have had their uncommitted changes leaked
**                           is now passed in, allowing proper cleanup.
** 008 ECF 20090720   @43310 Changed delete() method signature. Now it throws
**                           PersistenceException.
** 009 ECF 20140825          Track unvalidated inserts.
*/

package com.goldencode.p2j.persist.remote;

import java.io.*;
import java.util.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.util.*;

/**
 * Interface which defines a client-server API to manage the information about
 * uncommitted modifications. {@link RemoteDirtyShareManager}
 * invokes the methods of this interface on a local proxy for a remote object,
 * in order to access the dirty share services of the remote P2J
 * server which is authoritative for a particular database.
 * <p>
 * Since the remote P2J server may manage multiple databases, but only a
 * single instance of an implementor of this interface will be available at
 * any given remote server, dirty share service requests must be
 * multiplexed by database.  For this reason, each of the methods defined by
 * this interface accepts an integer ID, which uniquely identifies the target
 * database at the remote server.
 * <p>
 * The database ID to use for a particular physical database is determined by
 * calling {@link DirtyShareMultiplexer#getMultiplexID(Object)}. This is
 * the responsibility of the <code>RemoteDirtyShareManager</code> and must be
 * done before any dirty share management service is requested.
 * <p>
 * This interface effectively mirrors that of {@link
 * com.goldencode.p2j.persist.dirty.DirtyShareManager the DirtyShareManager
 * API}, except for the additional database multiplex identifier which is the
 * first parameter to each method.
 *
 * @author SVL
 */
public interface DirtyShareMultiplexer
extends RemoteMultiplexer<String>
{
   /**
    * Register with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given DMO entity.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          DMO entity for which the caller is interested in tracking
    *          change.
    * 
    * @return  A unique identifier for the registrant, assigned by the event
    *          clearinghouse, used later to collect events and deregister.
    * 
    * @see     #getGlobalEvents(int, long, long)
    * @see     #deregisterForGlobalEvents(int, long)
    */
   public long registerForGlobalEvents(int database, String entity);
   
   /**
    * Deregister with a global event clearinghouse an interest in insert,
    * delete, and update events which affect a particular DMO type.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   registerID
    *          A unique identifier for the registrant being deregistered.
    * 
    * @see     #getGlobalEvents(int, long, long)
    * @see     #registerForGlobalEvents(int, String)
    */
   public void deregisterForGlobalEvents(int database, long registerID);
   
   /**
    * Retrieve any new events that have been collected by a global event
    * clearinghouse, representing inserts, deletes, and updates made to
    * record types for which the caller previously has registered interest.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   registerID
    *          A unique identifier for an object which has registered interest
    *          in DMO changing events.
    * @param   lastID
    *          A unique identifier representing the last event retrieved by
    *          the caller.  If less than 0, it is assumed no events have yet
    *          been retrieved since the caller registered interest in these
    *          events.
    * 
    * @return  Array of applicable, new events, or <code>null</code> if no
    *          such events have been collected.
    * 
    * @throws  EventRegistrationException
    *          if <code>registerID</code> is invalid or has expired.
    */
   public GlobalChangeEvent[] getGlobalEvents(int database, long registerID, long lastID)
   throws EventRegistrationException;
   
   /**
    * Lock or unlock all indexes associated with the given DMO entity.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @throws  PersistenceException
    *          if there is an error querying JDBC metadata while retrieving
    *          index information.
    */
   public void lockAllIndexes(int database, String entity, int lockType)
   throws PersistenceException;
   
   /**
    * Begin tracking a newly inserted record in the dirty database.
    * <p>
    * This record is removed when the current transaction is committed or rolled back.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation class associated
    *          with a table being tracked for uncommitted changes.
    * @param   copy
    *          Record to be tracked.
    * @param   uniqueIndexes
    *          If the insertion is happening at record creation, and the record has not yet been
    *          validated, this will represent the collection of unique indexes that will need to
    *          be validated before the record can safely be persisted. If <code>null</code>, this
    *          indicates the record has been fully validated and should no longer be tracked as
    *          unvalidated.
    * @param   lock
    *          <code>true</code> to acquire write locks to insert the record; <code>false</code>
    *          to assume write locks are acquired by calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void insert(int database,
                      String entity,
                      Persistable copy,
                      Collection<String> uniqueIndexes,
                      boolean lock)
   throws PersistenceException;
   
   /**
    * Notify the dirty share manager that the record represented by the given key has been
    * validated, and should no longer be tracked as unvalidated. Refresh the data stored in the
    * dirty database with the potentially newer information in <code>copy</code>.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the remote server.
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * @param   copy
    *          Updated copy of DMO for the dirty manager to share with other contexts.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void notifyInsertValidated(int database, RecordIdentifier ident, Persistable copy)
   throws PersistenceException;
   
   /**
    * Indicate whether the record identified by <code>ident</code> represents a newly inserted
    * record, which has an unvalidated index, <code>index</code>. Assumes appropriate indexes
    * already are locked by the caller.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the remote server.
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * @param   index
    *          Index to be checked for validation.
    * 
    * @return  <code>true</code> if <code>ident</code> represents a new record in a foreign
    *          context with unvalidated index <code>index</code>.
    */
   public boolean isUnvalidatedIndex(int database, RecordIdentifier ident, String index);
   
   /**
    * Rollback the insert of a record previously introduced with the {@link
    * #insert(int, String, Persistable, Collection, boolean)} method.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   ident
    *          Identifier for the target record, which encapsulates its entity
    *          name and primary key.
    * @param   lock
    *          <code>true</code> to acquire write locks;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void rollbackInsert(int database, RecordIdentifier ident, boolean lock)
   throws PersistenceException;
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key is being tracked for changes within an uncommitted transaction.
    * This includes records which have been added to the dirty database due to
    * an insert or an update, not due to a delete.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> to indicate the target record is being
    *          tracked for changes, else <code>false</code>.
    */
   public boolean isTracked(int database, String entity, Serializable id);
   
   /**
    * Begin tracking a DMO for changes, based on the updates indicated by the
    * <code>properties</code> and <code>values</code> arguments.
    * <p>
    * Make a copy of the given record, apply the specified updates to that
    * copy, and store the copy in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   copy
    *          Record to be tracked.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   validatedIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(int database,
                      String entity,
                      Persistable copy,
                      String[] properties,
                      BaseDataType[] values,
                      Collection<String> validatedIndexes,
                      boolean lock)
   throws PersistenceException;
   
   /**
    * Retrieve a DMO which is being tracked in the dirty database, and apply
    * the specified updates to it, as provided by the <code>properties</code>
    * and <code>values</code> arguments.  It is assumed the DMO already exists
    * in the database, from a previous call to {@link
    * #update(int, String, Persistable, String[], BaseDataType[], Collection, boolean)}.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key id of the record to be retrieved and updated.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   validatedIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(int database,
                      String entity,
                      Serializable id,
                      String[] properties,
                      BaseDataType[] values,
                      Collection<String> validatedIndexes,
                      boolean lock)
   throws PersistenceException;
   
   /**
    * Overwrite the state of the records in the dirty database which match the
    * IDs of the given list of DMOs for the given entity, and store a snapshot
    * of each DMO in the map of newly inserted records.
    * <p>
    * This is necessary to support a quirk/bug in Progress which compromises
    * transaction isolation by leaking uncommitted changes (even non-indexed
    * data) under certain conditions.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   dmos
    *          Array of DMOs which are to be merged into their counterpart
    *          records in the dirty database and whose snapshots are to be
    *          stored in the map of newly inserted records.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void updateSnapshots(int database, String entity, Persistable[] dmos)
   throws PersistenceException;
   
   /**
    * Indicate whether any uncommitted modification is being tracked for the
    * database table associated with the specified entity.  A modification is
    * the insertion, deletion, or update of a record.  Insertions and updates
    * considered the same for purposes of this method.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * 
    * @return  Bitfield indicating dirty status of entity, using constants
    *          from {@link DirtyInfo}.  Zero return indicates entity is not
    *          dirty.  Only <code>CHANGE</code> and <code>DELETE</code> are
    *          detected, not <code>INSERT</code>.
    */
   public int getEntityDirtyState(int database, String entity);
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key has been deleted within an uncommitted transaction in <i>any</i>
    * context.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> if the specified record has been deleted,
    *          else <code>false</code>.
    */
   public boolean isDeleted(int database, String entity, Serializable id);
   
   /**
    * Mark a record as having been deleted in an uncommitted transaction, or
    * unmark a record which previously was so marked.
    * <p>
    * A record is unmarked when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of deleted record.
    * @param   rollback
    *          If <code>false</code>, the target record is marked as deleted;
    *          if <code>true</code>, it is unmarked.
    * @param   lock
    *          <code>true</code> to acquire write locks to delete the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there was an error deleting the record from the database.
    */
   public void delete(int database,
                      String entity,
                      Serializable id,
                      boolean rollback,
                      boolean lock)
   throws PersistenceException;
   
   /**
    * Retrieve from the dirty database the record specified by the given
    * entity and primary key, if it exists.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be retrieved.
    * 
    * @return  Target record if it exists in the dirty database, else
    *          <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public Persistable getDirtyDMO(int database, String entity, Serializable id)
   throws PersistenceException;
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   params
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(int database,
                    String hql,
                    Object[] params,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException;
   
   /**
    * Given certain search criteria and possibly the primary key of a
    * potential DMO match in the primary database, report any significant
    * information currently being tracked for uncommitted transactions, which
    * might override data (or the lack thereof) found in the primary database.
    * <p>
    * If a {@link DirtyInfo} object is returned by this method, it will
    * contain overriding information, such as a dirty DMO which matched the
    * search criteria, whether that record was newly inserted, or the fact
    * that the candidate record found in the primary search has been deleted
    * or modified in an uncommitted transaction.  A <code>null</code> return
    * indicates no overriding information was available.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   candidateID
    *          Primary key ID of DMO, if any, found in the primary database,
    *          which matches the current query criteria.
    * @param   index
    *          Name of index which governs the ordering of data in the current
    *          query.
    * @param   statements
    *          One or more HQL statements to be executed against the dirty
    *          database to find a record which matches the current query
    *          criteria.  The list is in order of most specific to least
    *          specific search criteria.
    * @param   params
    *          Query substitution parameters which match placeholders in the
    *          most specific HQL statement.
    * @param   types
    *          Hibernate data types of the query substitution parameters.
    * 
    * @return  An object which summarizes the findings, or <code>null</code>
    *          if no uncommitted change information was available for the
    *          given arguments.
    *          
    * @throws  PersistenceException
    *          if there was any database error.
    */
   public DirtyInfo getDirtyInfo(int database,
                                 String entity,
                                 Serializable candidateID,
                                 String index,
                                 List<String> statements,
                                 Object[] params,
                                 Type[] types)
   throws PersistenceException;
   
   /**
    * Stop tracking uncommitted changes to the the subset of records
    * represented by <code>subChanges</code>, and remove from the master set
    * of uncommitted, inserted records the subset of records represented by
    * <code>subInserts</code>.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   subChanges
    *          Sets of primary keys representing those records to be removed
    *          from uncommitted change tracking, mapped by entity name.
    * @param   subInserts
    *          A set of identifiers of those records to be removed from the
    *          master set of uncommitted inserts.
    * @param   leakedEntities
    *          Names of DMO types which have had their uncommitted updates
    *          "leaked" to other sessions via {@link #updateSnapshots(int,
    *          String, Persistable[])}.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void cleanupChanges(int database,
                              Map<String, Set<Serializable>> subChanges,
                              Set<RecordIdentifier> subInserts,
                              Set<String> leakedEntities)
   throws PersistenceException;
   
   /**
    * Remove from the master set of uncommitted, deleted records the subset of
    * records represented by <code>subDeletes</code>.
    * 
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   subDeletes
    *          Sets of primary keys representing those records to be removed
    *          from uncommitted delete tracking, mapped by entity name.
    */
   public void cleanupDeletes(int database, Map<String, Set<Serializable>> subDeletes);
}