/*
** Module   : TriggerTracker.java
** Abstract : Helper class to manage database trigger information for a single record buffer
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140801 Created initial version.
*/

package com.goldencode.p2j.persist.trigger;

import java.util.*;
import com.goldencode.p2j.persist.*;

/**
 * A helper class used by {@link RecordBuffer} to help manage information about database triggers.
 * This includes determining whether triggers of various types are registered for the buffer at
 * any given block, and tracking information about the execution and nesting of triggers.
 * <p>
 * Although the {@link DatabaseTriggerManager} manages much of the same information, it does so in
 * the context of determining whether or not to fire a given trigger, when such an event has been
 * requested. However, <code>RecordBuffer</code> must do some (in some cases non-trivial) setup
 * work in order to fire a trigger. This class provides a fast way to determine whether this work
 * is necessary (i.e., a trigger is registered in the current block), or can be avoided.
 * <p>
 * In order to properly prevent illegal nesting of trigger execution, this class provides a
 * mechanism to track which triggers currently are executing for the associated buffer. The
 * <code>RecordBuffer</code> class still does the "heavy lifting" of enforcing the rules, but the
 * data about trigger execution is managed here. This includes maintaining a stack of trigger
 * types currently executing, the names of fields currently involved in assign trigger execution,
 * and the names of fields whose assign triggers are currently deferred until the end of a batch
 * assign bracket.
 * <p>
 * Write triggers must be able to pass a DMO to the trigger procedure or block, which represents
 * a snapshot of the current record before the first field update was made. This snapshot is
 * created by <code>RecordBuffer</code>, but stored here for use when the trigger is fired. The
 * block depth at which that snapshot was created is stored as well, so that the snapshot can be
 * deleted in the event the block at that level is undone/rolled back.
 */
public final class TriggerTracker
{
   /** Constant indicating trigger enable status has not been checked */
   private static final int UNCHECKED = 0x0000;
   
   /** Database trigger manager */
   private DatabaseTriggerManager manager = DatabaseTriggerManager.get();
   
   /** Bit field indicating enable and check state for all trigger types for current scope */
   private int state = UNCHECKED;
   
   /** Class of record buffer using this helper */
   private final Class<? extends Buffer> bufferClass;
   
   /** DMO representing old buffer state for a write trigger */
   private Persistable oldDMO;
   
   /** Block level at which old DMO data was copied (when first assignment took place) */
   private int oldDMODepth;
   
   /** Map of property names to enable states for assign triggers */
   private Map<String, Boolean> assignTriggers = null;
   
   /** Stack of currently active trigger event types for the associated buffer */
   private Deque<DatabaseEventType> eventStack = null;
   
   /** Stack of names of fields for which assign triggers currently are executing */
   private Deque<String> fieldStack = null;
   
   /** Names of fields assigned in batch pending assign trigger events */
   private Set<String> pendingBatchFields = null;
   
   /**
    * Constructor.
    * 
    * @param   bufferClass
    *          Class of record buffer using this helper.
    */
   public TriggerTracker(Class<? extends Buffer> bufferClass)
   {
      this.bufferClass = bufferClass;
      resetOldDMO();
   }
   
   /**
    * Reset the enable and check state for all trigger types for the current scope.
    */
   public void resetState()
   {
      state = UNCHECKED;
   }
   
   /**
    * Indicate whether a trigger of the given event type is enabled for the associated buffer
    * (and property, in the case of an assign trigger), in the current scope. Caches information
    * to minimize lookups in the trigger manager.
    * 
    * @param   det
    *          Database event type.
    * @param   property
    *          Property name, in the case of an assign trigger (not used for other trigger types
    *          and may be <code>null</code>.
    * 
    * @return  <code>true</code> if a trigger of the given type is enabled, else
    *          <code>false</code>.
    */
   public boolean isTriggerEnabled(DatabaseEventType det, String property)
   {
      int enabledMask = det.getMask();
      int checkedMask = enabledMask << 8;
      
      // assign triggers require per-property handling, but can be disabled en masse
      if (DatabaseEventType.ASSIGN.equals(det))
      {
         if ((state & checkedMask) == UNCHECKED)
         {
            if (manager.areAssignTriggersEnabled(bufferClass))
            {
               // assign triggers enabled, but we don't yet know which are registered
               state |= enabledMask;
               assignTriggers = new HashMap<>();
            }
            else
            {
               assignTriggers = null;
            }
            
            // remember we checked enable status for this block
            state |= checkedMask;
         }
         
         if ((state & enabledMask) != enabledMask)
         {
            // all assign triggers are disabled
            return false;
         }
         
         Boolean enabled = assignTriggers.get(property);
         if (enabled == null)
         {
            enabled = manager.isAssignTrigger(bufferClass, property);
            assignTriggers.put(property, enabled);
         }
         
         return enabled;
      }
      
      // all other trigger types
      if ((state & checkedMask) == UNCHECKED)
      {
         // query state
         if (manager.isTriggerEnabled(det, bufferClass))
         {
            state |= enabledMask;
         }
         
         // remember that state has been checked in this scope
         state |= checkedMask;
      }
      
      return ((state & enabledMask) == enabledMask);
   }
   
   /**
    * Push a trigger type onto the stack of trigger types currently executing for the associated
    * record buffer. This should done just before the trigger of this type is fired.
    * 
    * @param   eventType
    *          Event type of the trigger about to be fired.
    */
   public void pushEvent(DatabaseEventType eventType)
   {
      if (eventStack == null)
      {
         eventStack = new ArrayDeque<>();
      }
      eventStack.push(eventType);
   }
   
   /**
    * Pop the most recent trigger type executed from the stack of trigger types currently
    * executing for the associated record buffer. This should be done unconditionally (i.e., in a
    * <code>finally</code> block), immediately after the trigger is fired.
    * 
    * @return  The trigger type popped from the top of the stack.
    * 
    * @throws  NoSuchElementException
    *          if the stack is empty (indicates a programming error).
    */
   public DatabaseEventType popEvent()
   {
      return eventStack.pop();
   }
   
   /**
    * Peek the most recent trigger type executed from the stack of trigger types currently
    * executing for the associated record buffer. This is safe to call if there is no trigger
    * currently executing (will return <code>null</code>).
    * 
    * @return  The trigger type most recently executed, or <code>null</code> if none.
    */
   public DatabaseEventType peekEvent()
   {
      return (eventStack == null ? null : eventStack.peek());
   }
   
   /**
    * Determine whether any trigger of the specified type currently is executing, in order to
    * allow the enforcement of recursion limitations for certain combinations of trigger types.
    * 
    * @param   eventType
    *          Trigger type to test.
    * 
    * @return  <code>true</code> if a trigger of the current type is executing for the associated
    *          record buffer (at any level), else <code>false</code>.
    */
   public boolean isExecuting(DatabaseEventType eventType)
   {
      return (eventStack != null && eventStack.contains(eventType));
   }
   
   /**
    * Push a field (DMO property) name for an assign trigger about to be executed, onto the stack
    * of field names of assign triggers currently executing for the associated record buffer.
    * 
    * @param   field
    *          DMO property name associated with an assign trigger.
    */
   public void pushField(String field)
   {
      if (fieldStack == null)
      {
         fieldStack = new ArrayDeque<>();
      }
      fieldStack.push(field);
   }
   
   /**
    * Pop the field (DMO property) name of the assign trigger most recently fired from the stack
    * of field names of assign triggers currently executing for the associated record buffer.
    * This should be done unconditionally (i.e., in a <code>finally</code> block), immediately
    * after the trigger is fired.
    * 
    * @return  The DMO property name popped from the top of the stack.
    * 
    * @throws  NoSuchElementException
    *          if the stack is empty (indicates a programming error).
    */
   public String popField()
   {
      return fieldStack.pop();
   }
   
   /**
    * Determine whether the specified field (DMO property) name represents an assign trigger which
    * currently is executing for the associated buffer, in order to allow the enforcement of
    * recursion limitations.
    * 
    * @param   field
    *          DMO property name associated with an assign trigger.
    * 
    * @return  <code>true</code> if an assign trigger for the given field is executing for the
    *          associated record buffer (at any level), else <code>false</code>.
    */
   public boolean isFieldInTrigger(String field)
   {
      return (fieldStack != null && fieldStack.contains(field));
   }
   
   /**
    * Determine whether the specified field (DMO property) name represents the assign trigger most
    * recently fired for the associated buffer, in order to allow the enforcement of recursion
    * limitations.
    * 
    * @param   field
    *          DMO property name associated with an assign trigger.
    * 
    * @return  <code>true</code> if an assign trigger for the given field is the most recently
    *          fired assign trigger for the associated record buffer, else <code>false</code>.
    */
   public boolean isFieldCurrentTrigger(String field)
   {
      if (fieldStack == null)
      {
         return false;
      }
      
      String top = fieldStack.peek();
      
      return field.equals(top);
   }
   
   /**
    * Add a field (DMO property) name to the set of fields whose assign triggers must be deferred
    * until the end of the current batch assign mode.
    * 
    * @param   field
    *          DMO property name whose assign trigger is pending.
    */
   public void addPendingBatchField(String field)
   {
      if (pendingBatchFields == null)
      {
         pendingBatchFields = new LinkedHashSet<>();
      }
      
      // remove the field first because we want to iterate in the order of last insertion
      pendingBatchFields.remove(field);
      pendingBatchFields.add(field);
   }
   
   /**
    * Get the set of field (DMO property) names whose assign triggers were deferred until the end
    * of the current batch assign mode.
    * 
    * @return  Set of DMO property names whose assign triggers are pending, or <code>null</code>,
    *          if none.
    */
   public Set<String> getPendingBatchFields()
   {
      return pendingBatchFields;
   }
   
   /**
    * Clear the set of field (DMO property) names whose assign triggers were deferred until the
    * end of the current batch assign mode.
    */
   public void clearPendingBatchFields()
   {
      pendingBatchFields = null;
   }
   
   /**
    * Get a DMO representing a snapshot of the old buffer state for a write trigger.
    * 
    * @return  Old DMO, may be <code>null</code>.
    */
   public Persistable getOldDMO()
   {
      return oldDMO;
   }
   
   /**
    * Set a DMO representing a snapshot of the old buffer state for a write trigger.
    * 
    * @param   oldDMO
    *          Old DMO; must not be <code>null</code>.
    * @param   blockLevel
    *          Block level at which <code>oldDMO</code> was copied from current record.
    * 
    * @throws  NullPointerException
    *          if <code>oldDMO</code> is <code>null</code>.
    */
   public void setOldDMO(Persistable oldDMO, int blockLevel)
   {
      if (oldDMO == null)
      {
         throw new NullPointerException("oldDMO must not be null");
      }
      
      this.oldDMO = oldDMO;
      this.oldDMODepth = blockLevel;
   }
   
   /**
    * Clear the snapshot of the most recently captured old buffer state for a write trigger and
    * forget the block level at which it was captured. This is done once the write trigger is
    * fired, or if the field assignments which caused the snapshot to be taken are rolled back.
    */
   public void resetOldDMO()
   {
      oldDMO = null;
      oldDMODepth = -1;
   }
   
   /**
    * If we are tracking an old DMO snapshot for a write trigger, and its current block or an
    * enclosing block is committed, track it at the depth of the current block's parent.
    * 
    * @param   blockDepth
    *          Current block depth of the buffer at the time of commit.
    */
   public void commit(int blockDepth)
   {
      if (blockDepth <= oldDMODepth)
      {
         oldDMODepth = blockDepth - 1;
      }
   }
   
   /**
    * If we are tracking an old DMO snapshot for a write trigger, and the block at which it is
    * being tracked is undone, reset the snapshot state.
    * 
    * @param   blockDepth
    *          Current block depth, which is compared with the block depth at which the snapshot
    *          (if any) is being tracked.
    */
   public void rollback(int blockDepth)
   {
      if (blockDepth == oldDMODepth)
      {
         resetOldDMO();
      }
   }
}
